/*
 * audio_ep7x12.c - Cirrus Logic EP7x1x DAI interface
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/tqueue.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/irq.h>
#include <asm/fiq.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <asm/arch/hardware.h>
#include <asm/hardware/ep7212.h>

#include <linux/soundcard.h>
#include <linux/interrupt.h>


#define DEBUG_DAI 0


#define DAI_MAJOR     14

#define DAI_FIQ 0

#if 0
#define DISABLE_DAI()   do { disable_fiq(DAI_FIQ); } while(0)
#define ENABLE_DAI()    do { enable_fiq(DAI_FIQ); } while(0)
#else
#define DISABLE_DAI()   do { cli(); IO_INTMR3=0; } while(0)
#define ENABLE_DAI()    do { IO_INTMR3=1; sti(); } while(0)
#endif


/* 
 * this is how many bytes at a time we can take.  4 samples, stereo, 16-bit (16 bytes).
 */
#define BUF_INCR    0x00000010
#define INCR_MASK   0x001FFFF0

#define BUF_SIZE    0x00200000
#define BUF_MASK    0x001FFFFF

/*
 * this is the max number of bytes that can be in the buffer.
 */
#define BUF_MAX     (BUF_SIZE - BUF_INCR)


static int dai_major;

char dai_buf[BUF_SIZE] __attribute__ ((aligned(BUF_INCR)));

volatile int dai_readp;
volatile int dai_writep;

int dai_wait_size=0;		/* block of free data that process is waiting to write */

/* list of waiting processes... */
DECLARE_WAIT_QUEUE_HEAD(dai_wait);

/* task for checking waiting processes and buffer level */
struct tq_struct dai_task;

#if DEBUG_DAI
/* task for checking underflow status... */
struct tq_struct dai_debug_task;
volatile int dai_underflow=0;
int dai_old_underflow=0;
void dai_debug_task_routine(void *data);
struct tq_struct dai_debug_task;
#endif 


/* which queue should we use */
#define DAI_TASK_QUEUE (&tq_timer)


void dai_task_routine(void *data);


/* struct fiq_handler { 
 * 	struct fiq_handler *next, 
 * 	char *name,
 * 	int (*fiq_op)(void *, int relinquish),
 * 	void *dev_id;
 */
static struct fiq_handler dai_fh = { NULL, "DAI", NULL, NULL };
extern unsigned char dai_start, dai_end;


#define SP_SIZE 32

unsigned char dai_sp[SP_SIZE] __attribute__ ((aligned(4)));;

volatile char *dai_flags;


static void dai_init(void)
{
	struct pt_regs regs;

	/*
	 * make sure FIQ is disabled first.
	 */
	DISABLE_DAI();

	printk(KERN_INFO "DAI: dai_init() initializing stuff\n");

	dai_flags = &dai_sp[SP_SIZE];
	memset(dai_buf, 0, BUF_SIZE);
	dai_readp = dai_writep = 0;

	get_fiq_regs(&regs);
	regs.ARM_r8	= (int) dai_buf;
	regs.ARM_r9	= 0;	/* scratch */
	regs.ARM_r10	= 0;	/* counts from zero to BUF_SIZE */
	//regs.ARM_fp	= 0;           /* r11 */
	regs.ARM_ip	= IO_BASE + DAIR;  /*0xff002000;*/	/* r12 */
	printk("DAI: regs.ARM_ip = 0x%08x\n", (unsigned int) regs.ARM_ip);
	regs.ARM_sp	= (int) dai_flags;
	set_fiq_regs(&regs);


	if (claim_fiq(&dai_fh)) {
		printk(KERN_WARNING "DAI: dai_init() couldn't claim FIQ.\n");
		return;
	}

	printk(KERN_INFO "DAI: dai_init() setting fiq handler\n");
	set_fiq_handler(&dai_start, &dai_end - &dai_start);

	/*
	 * Set up SYSCON3 and DAI64FS (if 7312)
	 */
	
	printk(KERN_INFO "DAI: dai_init() SYSCON1 (0x%08x)\n", (unsigned int) IO_SYSCON1);
	printk(KERN_INFO "DAI: dai_init() SYSCON2 (0x%08x)\n", (unsigned int) IO_SYSCON2);
	printk(KERN_INFO "DAI: dai_init() SYSCON3 (0x%08x)\n", (unsigned int) IO_SYSCON3);
	printk(KERN_INFO "DAI: dai_init() INTMR1  (0x%08x)\n", (unsigned int) IO_INTMR1);
	printk(KERN_INFO "DAI: dai_init() INTMR2  (0x%08x)\n", (unsigned int) IO_INTMR2);
	printk(KERN_INFO "DAI: dai_init() INTMR3  (0x%08x)\n", (unsigned int) IO_INTMR3);
	printk(KERN_INFO "DAI: dai_init() DAISR   (0x%08x)\n", (unsigned int) IO_DAISR);
	printk(KERN_INFO "DAI: dai_init() DAI64FS (0x%08X)\n", (unsigned int) IO_DAI64FS);

#define PHAT_64FS 0

#if PHAT_64FS
	IO_SYSCON3 |= SYSCON3_DAISEL;
	IO_SYSCON3 &= ~(1<<9);
	IO_DAI64FS |=	(2 << 8) | 
			DAI64FS_MCLK256EN | 
			DAI64FS_AUDCLKSRC | 
			DAI64FS_AUDCLKEN | 
			DAI64FS_I2SF64;
#else
	IO_SYSCON3 |= SYSCON3_DAISEL | SYSCON3_DAIEN;
#endif

	printk(KERN_INFO "DAI: dai_init() setting PE.1\n");

	/*
	 * Turn on the external 11.2896MHz clock
	 */
#define COPWR 1
	IO_PEDDR |= COPWR;
	IO_PEDR |= COPWR;
	
	printk(KERN_INFO "DAI: dai_init() setting DAI Control Register\n");
	/*
	 * DAI control register, 
	 * external clock source (ECS), 
	 * right channel transmit mask (RTCM).
	 */
	IO_DAIR = 0x0404 | DAIR_ECS | DAIR_RCTM; 

	/*
	 * Clear the overrun and underrun bits 
	 * in the DAI status register
	 */
	printk(KERN_INFO "DAI: dai_init() clearing DAI status register bits\n");
	IO_DAISR = 0xffffffff;


	printk(KERN_INFO "DAI: dai_init() setting DAIR_DAIEN\n");
	/*
	 * Enable DAI interface
	 */
	IO_DAIR |= DAIR_DAIEN;
	while (!(IO_DAISR & DAISR_FIFO)) {
	}

	/*
	 * Enable the FIFOs for the left and right channels.
	 *
	 * Left Channel.
	 */
	IO_DAIDR2 = DAIDR2_FIFOEN | DAIDR2_FIFOLEFT;
	while (!(IO_DAISR & DAISR_FIFO)) {
	}
	/*
	 * Right Channel.
	 */
	IO_DAIDR2 = DAIDR2_FIFOEN | DAIDR2_FIFORIGHT;
	while (!(IO_DAISR & DAISR_FIFO)) {
	}
	
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	IO_DAIDR1 = 0;
	IO_DAIDR0 = 0;
	
	printk(KERN_INFO "DAI: dai_init() DAISR = 0x%08x\n", (unsigned int) IO_DAISR);
	
	printk(KERN_INFO "DAI: dai_init() adding routine to task queue\n");
	dai_task.routine = dai_task_routine;
	dai_task.data = NULL;
	queue_task(&dai_task, DAI_TASK_QUEUE);


#if DEBUG_DAI
	printk(KERN_INFO "DAI: dai_init() adding debug task routine\n");
	dai_debug_task.routine = dai_debug_task_routine;
	dai_debug_task.data = NULL;
	queue_task(&dai_debug_task, DAI_TASK_QUEUE);
#endif

	/*
	 * unmask the DAI interrupt
	 */
	printk(KERN_INFO "DAI: dai_init() enabling DAI interrupt\n");

	/*
	 * Enable bit in INTMR3
	 */
	ENABLE_DAI();
}

#if DEBUG_DAI
void dai_debug_task_routine(void *data)
{
	if (dai_underflow && !dai_old_underflow) {
		printk("DAI: buffer underflow!  skip dectected!\n");
	}

	dai_old_underflow = dai_underflow;

	/* enqueue ourselves again */
	queue_task(&dai_debug_task, DAI_TASK_QUEUE);
}

#endif



void dai_task_routine(void *data)
{
	int readp;
	int avail;

	if (dai_wait_size) {
		readp = dai_readp;
		if (readp == dai_writep) {
			/* empty case */
			avail = BUF_MAX;
		}
		else if (readp < dai_writep) {
			avail = BUF_MAX - dai_writep + readp;
		}
		else {
			avail = readp - dai_writep - BUF_INCR;
		}
		if (avail >= dai_wait_size) {
			/*
			 * wake up waiting process.
			 */
			dai_wait_size = 0;
			wake_up_interruptible(&dai_wait);
		}
	}
	/* enqueue ourselves again */
	queue_task(&dai_task, DAI_TASK_QUEUE);
}


static ssize_t dai_write(struct file *file,
			 const char *buffer, size_t count, loff_t * loff)
{
	int to_copy;
	int this_copy;
	int readp;
	int avail;

	//printk(KERN_DEBUG "dai: write(count = %d)\n", count);

	/*
	 * should always always always write a multiple of 4 stereo 16 bit samples
	 * (i.e., 16 bytes).
	 */

	count &= INCR_MASK;
	
	/* 
	 * Figure out how much space we have available
	 */

dai_write_test_count:
	readp = dai_readp;
	if (readp == dai_writep) {
		/* empty case */
		avail = BUF_MAX;
	}
	else if (readp < dai_writep) {
		avail = BUF_MAX - dai_writep + readp;
	}
	else {
		avail = readp - dai_writep - BUF_INCR;
	}

	/*
	 * if we can't write the whole chunk, then tell the app to try again later.
	 */
	if (avail < count) {
		if (file->f_flags & O_NONBLOCK) 
			return -EAGAIN;
		else {
			/*
			 * put process to sleep.
			 */
			dai_wait_size = count;
			interruptible_sleep_on(&dai_wait);
			goto dai_write_test_count;
		}
	}

	/*
	 * copy.
	 */
	to_copy = count;

	while (to_copy > 0) {
		if ((dai_writep + to_copy) > BUF_SIZE)
			this_copy = BUF_SIZE - dai_writep;
		else
			this_copy = to_copy;

		copy_from_user(dai_buf + dai_writep, buffer, this_copy);

		buffer += this_copy;
		to_copy -= this_copy;

		/* This is the critical part. */
		DISABLE_DAI();
#if DEBUG_DAI
		dai_underflow=0;
#endif
		dai_writep += this_copy;
		if (dai_writep >= BUF_SIZE)
			dai_writep = 0;
		ENABLE_DAI();
	}
	return count;
}


static int dai_open(struct inode *inode, struct file *file)
{
	//printk(KERN_DEBUG "dai: open()\n");
	return 0;
}

#if PHAT_USE_WAIT_QUEUE
static int dai_flush(struct inode *inode, struct file *file)
{
	//interruptible_sleep_on(&flush_queue);

	return wait_event_interruptible(flush_queue,
					(dai_readp == dai_writep));
}
#endif



static int dai_release(struct inode *inode, struct file *file)
{
	//printk(KERN_DEBUG "dai: release()\n");
	return 0;
}


static void dai_clear(struct inode *inode, struct file *file)
{
	//printk(KERN_DEBUG "dai: clear()\n");
	short *daidata;

	DISABLE_DAI();
	/*
	 * This clears the buffer of pending data.
	 */
	dai_writep = dai_readp;

	/*
	 * The following code is to "smooth out" the discontinuity created in the output.
	 */
	daidata = (short *) &dai_buf[dai_writep];
	daidata[0] /= 2;
	daidata[1] /= 2;
	daidata[2] = daidata[0] / 2;
	daidata[3] = daidata[1] / 2;
	daidata[4] = daidata[2] / 2;
	daidata[5] = daidata[3] / 2;
	daidata[6] = daidata[4] / 2;
	daidata[7] = daidata[5] / 2;
	dai_writep += 16;
	if (dai_writep >= BUF_SIZE)
		dai_writep = 0;

	ENABLE_DAI();
}



static int dai_ioctl(struct inode *inode,
		     struct file *file, uint command, ulong arg)
{
	int val;
	switch (command) {
	case SNDCTL_DSP_GETFMTS:
		put_user(AFMT_U16_LE, (long *) arg);
		break;
	case SNDCTL_DSP_SPEED:
		/*
		 * Use this to set 64fs mode.
		 */
		if (get_user(val, (int *)arg))
			return -EFAULT;

		DISABLE_DAI();
		IO_SYSCON3 |= SYSCON3_DAISEL;
		IO_SYSCON3 &= ~(1<<9);
		IO_DAI64FS |=	(2 << 8) | DAI64FS_MCLK256EN | 
			DAI64FS_AUDCLKSRC | DAI64FS_AUDCLKEN | 
			DAI64FS_I2SF64;
		ENABLE_DAI();
		break;
	case SNDCTL_DSP_STEREO:
	case SNDCTL_DSP_SETFMT:
		break;
	case SNDCTL_DSP_RESET:
		/*
		 * clear buffers immediately.
		 */
		dai_clear(inode, file);
		break;
	case SNDCTL_DSP_SYNC:
		/*
		 * Send a value to the user.  
		 *
		 * 0 if buffer is empty
		 * 1 if buffer contains data.
		 */
		if (get_user(val, (long *) arg)) val = 0;
		if (val) {
		  if (dai_readp < dai_writep) {
		    val = (dai_readp + val) < dai_writep;
		  } else if (dai_readp > dai_writep) {
		    val = (dai_readp + val) < (dai_writep+BUF_SIZE);
		  } else {
		    val = 0;
		  }
		} else {
		  val = dai_readp != dai_writep;
		}
		put_user(val, (long *) arg);
		break;
	case SNDCTL_DSP_GETBLKSIZE:
		put_user(BUF_SIZE, (long *) arg);
		break;
	case SNDCTL_DSP_SETFRAGMENT:
		get_user(val, (long *) arg);
		break;
	case SOUND_MIXER_READ_VOLUME:
		put_user(128, (long *) arg);
		break;
	case SOUND_MIXER_READ_PCM:
		put_user(128, (long *) arg);
		break;
	case SNDCTL_DSP_GETCAPS:
		put_user(DSP_CAP_REALTIME, (long *) arg);
		break;
	case SNDCTL_DSP_GETOSPACE:
		if (dai_readp == dai_writep) val = 0;
		else if (dai_readp < dai_writep) val = dai_writep-dai_readp;
		else val = dai_writep+BUF_SIZE-dai_readp;
		val = BUF_SIZE - val;
		put_user(val, (long *) arg);
		break;
	default:
		return 0;
		//return -EINVAL;
	}
	/*
	 * 0 == all good.
	 */
	return 0;
}


struct file_operations dai_fops = {
	write:		dai_write,
	ioctl:		dai_ioctl,
	open:		dai_open,
	release:	dai_release,
};


static int __init audio_7110_init(void) 
{
	printk(KERN_INFO "DAI: Version 1.2\n");

	dai_major = register_chrdev(DAI_MAJOR, "dai", &dai_fops);

	if (dai_major) {
		printk(KERN_WARNING "DAI: unable to get major %d\n", DAI_MAJOR);
	}
	else {
		dai_major = DAI_MAJOR;
		printk(KERN_INFO "DAI: major %d\n", dai_major);
		dai_init();
	}
	return 0;
}


module_init(audio_7110_init);

